@@ -32,23 +32,29 @@ import tempfile
 import shutil
 import numpy
 import pandas
+import json
 import xarray
 import cdflib
+import netCDF4
 from ._wps import time_util
 if os.name == "nt":
     import atexit
 
+from ._data import CONFIG_SWARM
+from ._data import CONFIG_AEOLUS
+
 CDF_EPOCH_1970 = 62167219200000.0
 
+ALLOWED_SPACECRFTS = ["A", "B", "C", "1", "2", "-"]
+
 # Frame names to use as xarray dimension names
 FRAME_NAMES = {
-    "NEC": ["B_NEC"],
+    "NEC": ["B_NEC", "B_OB", "B_CF", "B_SV", "sigma_OB", "sigma_CF", "sigma_SV"],
     "VFM": ["B_VFM", "dB_Sun", "dB_AOCS", "dB_other", "B_error"],
     "quaternion": ["q_NEC_CRF"],
     "WGS84": ["GPS_Position", "LEO_Position"],
     "EEJ_QDLat": ["EEJ"],
     "NE": ["J_NE", "J_CF_NE", "J_DF_NE", "B_NE"],
-    "AOB_Quality": ["Quality"],
 }
 # Reverse mapping of the above
 DATANAMES_TO_FRAME_NAMES = {}
@@ -64,14 +70,12 @@ FRAME_LABELS = {
     "WGS84": ["X", "Y", "Z"],
     "EEJ_QDLat": numpy.linspace(-20, 20, 81),
     "NE": ["N", "E"],
-    "AOB_Quality": ["Pa", "Sigma"],
 }
 FRAME_DESCRIPTIONS = {
     "NEC": "NEC frame - North, East, Centre (down)",
     "NE": "Horizontal NE frame - North, East",
     "VFM": "Vector Field Magnetometer instrument frame",
     "EEJ_QDLat": "Quasi-dipole latitude profile between -20 and 20 degrees from the EEF product",
-    "AEJ_Quality": "Boundary quality indicator (Pa, Sigma)",
 }
 
 
@@ -137,10 +141,10 @@ class FileReader(object):
         return data
 
     def get_variable_units(self, var):
-        return self._varatts[var].get("UNITS", None)
+        return self._varatts[var].get("UNITS", "")
 
     def get_variable_description(self, var):
-        return self._varatts[var].get("DESCRIPTION", None)
+        return self._varatts[var].get("DESCRIPTION", "")
 
     def get_variable_numdims(self, var):
         return self._varinfo[var].get("Num_Dims")
@@ -206,7 +210,7 @@ class FileReader(object):
                 df[column + "_" + str(suffix)] = vector_data[:, i]
         return df
 
-    def as_xarray_dataset(self):
+    def as_xarray_dataset(self, reshape=False):
         # NB currrently does not set the global metadata (attrs)
         #  (avoids issues with concatenating them)
         #  (this is done in ReturnedData)
@@ -215,11 +219,10 @@ class FileReader(object):
             coords={"Timestamp":
                     self._cdftime_to_datetime(self.get_variable("Timestamp"))})
         # Add Spacecraft variable as Categorical to save memory
-        ds["Spacecraft"] = (("Timestamp",), pandas.Categorical(
-            self.get_variable("Spacecraft"), categories=["A", "B", "C", "-"]))
-        datanames = set(self.variables)
-        datanames.remove("Timestamp")
-        datanames.remove("Spacecraft")
+        if "Spacecraft" in self.variables:
+            ds["Spacecraft"] = (("Timestamp",), pandas.Categorical(
+                self.get_variable("Spacecraft"), categories=ALLOWED_SPACECRFTS))
+        datanames = set(self.variables) - {"Timestamp", "Spacecraft"}
         # Loop through each variable available and append them to the Dataset,
         #  attaching the Timestamp coordinate to each.
         # Attach dimension names based on the name of the variable,
@@ -256,16 +259,90 @@ class FileReader(object):
         for dimname, dimlabels in FRAME_LABELS.items():
             if dimname in dims_used:
                 ds[dimname] = numpy.array(dimlabels)
-                ds[dimname].attrs["description"] = FRAME_DESCRIPTIONS.get(
-                    dimname, None)
                 ds = ds.set_coords(dimname)
+        #         ds[dimname].attrs["description"] = FRAME_DESCRIPTIONS.get(
+        #             dimname, None)
+        #         ds = ds.set_coords(dimname)
+        # Reshape to a sensible higher dimensional structure
+        # Currently only for GVO data, and without magnetic model values or auxiliaries
+        # Inefficient as it is duplicating the data (ds -> ds2)
+        if reshape:
+            ds = self.reshape_dataset(ds)
         # Add metadata of each variable
-        for dataname in ds:
-            ds[dataname].attrs["units"] = self.get_variable_units(dataname)
-            ds[dataname].attrs["description"] = self.get_variable_description(
-                dataname)
+        for var in list(ds.data_vars) + list(ds.coords):
+            try:
+                ds[var].attrs["units"] = self.get_variable_units(var)
+            except KeyError:
+                ds[var].attrs["units"] = ""
+            try:
+                ds[var].attrs["description"] = self.get_variable_description(var)
+            except KeyError:
+                ds[var].attrs["description"] = FRAME_DESCRIPTIONS.get(var, "")
+        # Remove unused Timestamp unit (-)
+        # for xarray 0.17 compatibility when writing to netcdf
+        ds["Timestamp"].attrs.pop("units", None)
         return ds
 
+    @staticmethod
+    def reshape_dataset(ds):
+        if "SiteCode" in ds.data_vars:
+            codevar = "SiteCode"
+        elif "IAGA_code" in ds.data_vars:
+            codevar = "IAGA_code"
+        else:
+            raise NotImplementedError(
+                """
+                Only available for GVO dataset where the "SiteCode"
+                parameter has been requested, or OBS dataset with "IAGA_code"
+                """
+            )
+        # Create integer "Site" identifier based on SiteCode / IAGA_code
+        sites = dict(enumerate(sorted(set(ds[codevar].values))))
+        sites_inv = {v: k for k, v in sites.items()}
+        if len(sites) == 0:
+            _ds_locs = ds
+        else:
+            # Identify (V)OBS locations and mapping from integer "Site" identifier
+            pos_vars = ["Longitude", "Latitude", "Radius", codevar]
+            _ds_locs = next(iter(ds[pos_vars].groupby("Timestamp")))[1]
+            if len(sites) > 1:
+                _ds_locs = _ds_locs.drop(("Timestamp")).rename({"Timestamp": "Site"})
+            else:
+                _ds_locs = _ds_locs.drop(("Timestamp")).expand_dims("Site")
+            _ds_locs["Site"] = [sites_inv.get(code) for code in _ds_locs[codevar].values]
+            _ds_locs = _ds_locs.sortby("Site")
+        # Create dataset initialised with the (V)OBS positional info as coords
+        # and datavars (empty) reshaped to (Site, Timestamp, ...)
+        t = numpy.unique(ds["Timestamp"])
+        ds2 = xarray.Dataset(
+            coords={
+                "Timestamp": t,
+                codevar: (("Site"), _ds_locs[codevar].data),
+                "Latitude": ("Site", _ds_locs["Latitude"].data),
+                "Longitude": ("Site", _ds_locs["Longitude"].data),
+                "Radius": ("Site", _ds_locs["Radius"].data),
+                "NEC": ["N", "E", "C"]
+            },
+        )
+        # (Dropping unused Spacecraft var)
+        data_vars = set(ds.data_vars) - {"Latitude", "Longitude", "Radius", codevar, "Spacecraft"}
+        N_sites = len(_ds_locs[codevar])
+        # Create empty data variables to be infilled
+        for var in data_vars:
+            shape = [N_sites, len(t), *ds[var].shape[1:]]
+            ds2[var] = ("Site", *ds[var].dims), numpy.empty(shape, dtype=ds[var].dtype)
+            ds2[var][...] = None
+        # Loop through each (V)OBS site to infill the data
+        if N_sites != 0:
+            for k, _ds in dict(ds.groupby(codevar)).items():
+                site = sites_inv.get(k)
+                for var in data_vars:
+                    ds2[var][site, ...] = _ds[var].values
+        # Revert to using only the "SiteCode"/"IAGA_code" identifier
+        ds2 = ds2.set_index({"Site": codevar})
+        ds2 = ds2.rename({"Site": codevar})
+        return ds2
+
 
 def make_pandas_DataFrame_from_csv(csv_filename):
     """Load a csv file into a pandas.DataFrame
@@ -336,7 +413,7 @@ class ReturnedDataFile(object):
     def open_cdf(self):
         """Returns the opened file as cdflib.CDF
         """
-        return FileReader.open_cdf(self._file.name)
+        return FileReader._open_cdf(self._file.name)
 
     def _write_new_data(self, data):
         """Replace the tempfile contents with 'data' (bytes)
@@ -427,12 +504,14 @@ class ReturnedDataFile(object):
             if expand:
                 raise NotImplementedError
             df = make_pandas_DataFrame_from_csv(self._file.name)
+        elif self.filetype == 'nc':
+            df = self.as_xarray().to_dataframe()
         elif self.filetype == 'cdf':
             with FileReader(self._file) as f:
                 df = f.as_pandas_dataframe(expand=expand)
         return df
 
-    def as_xarray(self, group=None):
+    def as_xarray(self, group=None, reshape=False):
         """Convert the data to an xarray Dataset.
 
         Note:
@@ -448,15 +527,86 @@ class ReturnedDataFile(object):
             raise NotImplementedError("csv to xarray is not supported")
         elif self.filetype == 'cdf':
             with FileReader(self._file) as f:
-                ds = f.as_xarray_dataset()
+                ds = f.as_xarray_dataset(reshape=reshape)
         elif self.filetype == 'nc':
-            ds = xarray.open_dataset(self._file.name, group=group)
+            # xarrays open_dataset does not retrieve data in groups
+            # group needs to be specified while opening
+            # we iterate here over the available groups
+            # TODO: what happens with groups of different sizes and attributes
+            nc = netCDF4.Dataset(self._file.name)
+            ds = xarray.Dataset()
+
+            # some datasets do not have groups
+            if nc.groups:
+                for group in nc.groups:
+                    ds = ds.merge(xarray.open_dataset(
+                        self._file.name, group=group, engine='netcdf4'
+                    ))
+            else:
+                ds = xarray.open_dataset(
+                    self._file.name, engine='netcdf4'
+                )
+            # Go through Aeolus parameters and check if unit information is available
+            # TODO: We are "flattening" the list of parameters, same parameter
+            # id in different collection types could select incorrect one
+            for parameter in ds:
+                for coll_obj in CONFIG_AEOLUS["collections"].values():
+                    for field_type in coll_obj.values():
+                        if parameter in field_type and field_type[parameter]['uom']:
+                            ds[parameter].attrs["units"] = field_type[parameter]['uom']
+            # TODO: Go through Swarm parameters
         return ds
 
+    def as_xarray_dict(self):
+        """Convert the data to an xarray Dataset.
+
+        Note:
+            Only supports netCDF format
+
+        Returns:
+            dict of xarray.Dataset
+
+        """
+        if self.filetype == 'csv':
+            raise NotImplementedError("csv to xarray dict is not supported")
+        elif self.filetype == 'cdf':
+            raise NotImplementedError("cdf to xarray dict is not supported")
+        elif self.filetype == 'nc':
+            result_dict = {}
+            nc = netCDF4.Dataset(self._file.name)
+            # some datasets do not have groups
+            if nc.groups:
+                for group in nc.groups:
+                    ds = xarray.Dataset()
+                    ds = ds.merge(xarray.open_dataset(
+                        self._file.name, group=group, engine='netcdf4'
+                    ))
+                    for parameter in ds:
+                        for coll_obj in CONFIG_AEOLUS["collections"].values():
+                            for field_type in coll_obj.values():
+                                if parameter in field_type and field_type[parameter]['uom']:
+                                    ds[parameter].attrs["units"] = field_type[parameter]['uom']
+                    result_dict[group] = ds
+            else:
+                result_dict['group'] = xarray.open_dataset(
+                    self._file.name, engine='netcdf4'
+                )
+            
+        return result_dict
+
     @property
     def sources(self):
-        with FileReader(self._file) as f:
-            sources = f.sources
+        if self.filetype == "nc":
+            nc = netCDF4.Dataset(self._file.name)
+            json_hist = json.loads(nc.history)
+            sources = [elem for elem in zip(
+                json_hist["inputFiles"],
+                json_hist["baselines"],
+                json_hist["software_vers"],
+            )]
+        else:
+            with FileReader(self._file) as f:
+                sources = f.sources
         return sources
 
     @property
@@ -485,6 +635,7 @@ class ReturnedData(object):
         data.range_filters
         data.magnetic_models
         data.as_xarray()
+        data.as_xarray_dict()
         data.as_dataframe(expand=True)
         data.to_file()
 
@@ -577,9 +728,12 @@ class ReturnedData(object):
         return pandas.concat(
             [d.as_dataframe(expand=expand) for d in self.contents])
 
-    def as_xarray(self):
+    def as_xarray(self, reshape=False):
         """Convert the data to an xarray Dataset.
 
+        Args:
+            reshape (bool): Reshape to a convenient higher dimensional form
+
         Returns:
             xarray.Dataset
 
@@ -590,22 +744,32 @@ class ReturnedData(object):
         #  and the filtering that has been applied.
         ds_list = []
         for i, data in enumerate(self.contents):
-            ds_part = data.as_xarray()
+            ds_part = data.as_xarray(reshape=reshape)
             if ds_part is None:
                 print("Warning: ",
-                      "Unable to create dataset from part {} of {}".format(
+                    "Unable to create dataset from part {} of {}".format(
                         i+1, len(self.contents)),
-                      "\n(This part is likely empty)")
+                    "\n(This part is likely empty)")
             else:
-                # Collect the non-empty Datasets
                 ds_list.append(ds_part)
-        if len(ds_list) == 1:
+        ds_list = [i for i in ds_list if i is not None]
+
+        if ds_list == []:
+            return None
+        elif len(ds_list) == 1:
             ds = ds_list[0]
         else:
-            ds_list = [i for i in ds_list if i is not None]
-            if ds_list == []:
+            dims = [d for d in list(ds_list[0].dims) if "array" not in d]
+            if dims == []:
                 return None
-            ds = xarray.concat(ds_list, dim="Timestamp")
+            elif len(dims) == 1:
+                ds = xarray.concat(ds_list, dim=dims[0])
+            else:
+                ds_list_per_dim = []
+                for d in dims:
+                    drop_dims = [dd for dd in dims if dd != d]
+                    ds_list_per_dim.append(xarray.concat([_ds.drop_dims(drop_dims) for _ds in ds_list], dim=d))
+                ds = xarray.merge(ds_list_per_dim)
         # # Test this other option:
         # ds = self.contents[0].as_xarray()
         # for d in self.contents[1:]:
@@ -616,11 +780,53 @@ class ReturnedData(object):
         # concat is slow. Maybe try extracting numpy arrays and rebuilding ds
 
         # Set the original data sources and models used as metadata
+        # only for cdf data types
         ds.attrs["Sources"] = self.sources
-        ds.attrs["MagneticModels"] = self.magnetic_models
-        ds.attrs["RangeFilters"] = self.range_filters
+        if self.filetype == "cdf":
+            ds.attrs["MagneticModels"] = self.magnetic_models
+            ds.attrs["RangeFilters"] = self.range_filters
         return ds
 
+    def as_xarray_dict(self):
+        """Convert the data to a dict containing an xarray per group.
+
+        Returns:
+            dict of xarray.Dataset
+
+        """
+        # ds_list is a list of xarray.Dataset objects
+        #  - they are created from each file in self.contents
+        # Some of them may be empty because of the time window they cover
+        #  and the filtering that has been applied.
+        ds_list = []
+        for i, data in enumerate(self.contents):
+            ds_part = data.as_xarray_dict()
+            if ds_part is None:
+                print("Warning: ",
+                    "Unable to create dataset from part {} of {}".format(
+                        i+1, len(self.contents)),
+                    "\n(This part is likely empty)")
+            else:
+                ds_list.append(ds_part)
+        ds_list = [i for i in ds_list if i is not None]
+        if ds_list == []:
+            return None
+        elif len(ds_list) == 1:
+            # add sources to all dict as_xarray
+            for xa_ds in ds_list[0].values():
+                xa_ds.attrs["Sources"] = self.sources
+            ds_dict = ds_list[0]
+        else:
+            ds_dict = {}
+            ds_list = [i for i in ds_list if i is not None]
+            if ds_list == []:
+                return None
+            for group in ds_list[0]:
+                group_list = [g[group] for g in ds_list if g is not None]
+                ds_dict[group] = xarray.merge(group_list)
+                ds_dict[group].attrs["Sources"] = self.sources
+        return ds_dict
+
     def to_files(self, paths, overwrite=False):
         """Saves the data to the specified files.
 
