@@ -33,9 +33,10 @@ import logging
 from argparse import ArgumentParser
 from .common import Command
 from .configuration import (
-    SetTokenCommand, SetPasswordCommand, RemoveServerCommand,
+    SetTokenCommand, RemoveServerCommand,
     SetDefaultServerCommand, RemoveDefaultServerCommand,
     ShowConfigurationCommand, InitializeConfigurationCommand,
+    ClearCredentialsCommand
 )
 from .upload import (
     UploadDataFileCommand, ShowUploadsCommand, RemoveUploadsCommand,
@@ -46,11 +47,11 @@ from .upload import (
 COMMANDS = {
     # configuration commands
     "set_token": SetTokenCommand(),
-    "set_password": SetPasswordCommand(),
     "remove_server": RemoveServerCommand(),
     "set_default_server": SetDefaultServerCommand(),
     "remove_default_server": RemoveDefaultServerCommand(),
     "show_configuration": ShowConfigurationCommand(),
+    "clear_credentials": ClearCredentialsCommand(),
     # data upload commands
     "upload_file": UploadDataFileCommand(),
     "show_uploads": ShowUploadsCommand(),
